@@ -1,11 +1,13 @@
 """Action to support type hints."""
 
+import copy
 import inspect
 import os
 import re
+import typing
 import warnings
 import yaml
-from argparse import Action, Namespace
+from argparse import Action
 from collections.abc import Iterable as abcIterable
 from collections.abc import Mapping as abcMapping
 from collections.abc import MutableMapping as abcMutableMapping
@@ -32,38 +34,34 @@ from typing import (
     Union,
 )
 
-try:
-    from typing import Literal  # type: ignore
-except ImportError:
-    Literal = False  # type: ignore
-
-from .actions import _is_action_value_list
+from .actions import _find_action, _is_action_value_list
+from .namespace import is_empty_namespace, Namespace
 from .typing import get_import_path, is_final_class, object_path_serializer, registered_types
 from .optionals import (
     argcomplete_warn_redraw_prompt,
     files_completer,
-    ModuleNotFound,
 )
 from .util import (
     change_to_path_dir,
     import_object,
     ParserError,
     Path,
-    namespace_to_dict,
-    get_key_value_from_flat_dict,
-    update_key_value_in_flat_dict,
     NoneType,
     yamlParserError,
     yamlScannerError,
     indent_text,
     _issubclass,
     _load_config,
+    warning,
 )
 
 
 __all__ = ['ActionTypeHint', 'lazy_instance']
 
 
+Literal = getattr(typing, 'Literal', False)
+
+
 root_types = {
     bool,
     Any,
@@ -84,9 +82,14 @@ leaf_types = {
     NoneType,
 }
 
-not_subclass_types = set(k for k in registered_types.keys() if not isinstance(k, tuple))  # type: Set
+not_subclass_types: Set = set(k for k in registered_types.keys() if not isinstance(k, tuple))
 not_subclass_types = not_subclass_types.union(leaf_types).union(root_types)
 
+tuple_set_origin_types = {Tuple, tuple, Set, set, frozenset, MutableSet, abcSet, abcMutableSet}
+sequence_origin_types = {List, list, Iterable, Sequence, MutableSequence, abcIterable, abcSequence,
+                         abcMutableSequence}
+mapping_origin_types = {Dict, dict, Mapping, MutableMapping, abcMapping, abcMutableMapping}
+
 
 class ActionTypeHint(Action):
     """Action to parse a type hint."""
@@ -108,7 +111,7 @@ class ActionTypeHint(Action):
         """
         if typehint is not None:
             if not self.is_supported_typehint(typehint, full=True):
-                raise ValueError('Unsupported type hint '+str(typehint)+'.')
+                raise ValueError(f'Unsupported type hint {typehint}.')
             self._typehint = typehint
             self._enable_path = False if is_optional(typehint, Path) else enable_path
         elif '_typehint' not in kwargs:
@@ -143,8 +146,7 @@ class ActionTypeHint(Action):
 
     @staticmethod
     def is_class_typehint(typehint, only_subclasses=False):
-        if isinstance(typehint, Action):
-            typehint = getattr(typehint, '_typehint', None)
+        typehint = typehint_from_action(typehint)
         typehint_origin = getattr(typehint, '__origin__', None)
         if typehint_origin == Union:
             subtypes = [a for a in typehint.__args__ if a != NoneType]
@@ -154,11 +156,52 @@ class ActionTypeHint(Action):
         return inspect.isclass(typehint) and typehint not in not_subclass_types and typehint_origin is None
 
 
+    @staticmethod
+    def is_final_class_typehint(typehint):
+        typehint = typehint_from_action(typehint)
+        typehint_origin = getattr(typehint, '__origin__', None)
+        if typehint_origin == Union:
+            subtypes = [a for a in typehint.__args__ if a != NoneType]
+            if len(subtypes) == 1:
+                typehint = subtypes[0]
+        return is_final_class(typehint)
+
+
     @staticmethod
     def is_subclass_typehint(typehint):
         return ActionTypeHint.is_class_typehint(typehint, True)
 
 
+    @staticmethod
+    def is_mapping_class_typehint(typehint):
+        typehint = typehint_from_action(typehint)
+        typehint_origin = getattr(typehint, '__origin__', None)
+        if typehint_origin not in mapping_origin_types:
+            return False
+        return ActionTypeHint.is_class_typehint(getattr(typehint, '__args__')[1])
+
+
+    @staticmethod
+    def is_mapping_typehint(typehint):
+        typehint_origin = getattr(typehint, '__origin__', typehint)
+        if typehint in mapping_origin_types or typehint_origin in mapping_origin_types or is_optional(typehint, tuple(mapping_origin_types)):
+            return True
+        return False
+
+
+    def is_init_arg_mapping_typehint(self, key, cfg):
+        result = False
+        class_path = cfg.get(f'{self.dest}.class_path')
+        if isinstance(class_path, str) and key.startswith(f'{self.dest}.init_args.') and self.is_subclass_typehint(self):
+            sub_add_kwargs = dict(self.sub_add_kwargs)
+            sub_add_kwargs.pop('linked_targets', None)
+            parser = ActionTypeHint.get_class_parser(class_path, sub_add_kwargs=sub_add_kwargs)
+            key = re.sub(f'^{self.dest}.init_args.', '', key)
+            typehint = getattr(_find_action(parser, key), '_typehint', None)
+            result = self.is_mapping_typehint(typehint)
+        return result
+
+
     @staticmethod
     def parse_subclass_arg(parser, arg_string):
         if '.class_path' in arg_string or '.init_args.' in arg_string:
@@ -166,7 +209,7 @@ class ActionTypeHint(Action):
                 arg_base, explicit_arg = arg_string.rsplit('.class_path', 1)
             else:
                 arg_base, init_arg = arg_string.rsplit('.init_args.', 1)
-                match = re.match(r'([\w_]+)(|=.*)$', init_arg)
+                match = re.match(r'(\w+)(|=.*)$', init_arg)
                 if match:
                     explicit_arg = match.groups()[1]
             action = parser._option_string_actions.get(arg_base)
@@ -204,29 +247,24 @@ class ActionTypeHint(Action):
             cfg, val, opt_str = args[1:]
             if isinstance(opt_str, str):
                 if opt_str.endswith('.class_path'):
-                    cfg_dest = get_key_value_from_flat_dict(vars(cfg), self.dest)
+                    cfg_dest = cfg.get(self.dest, Namespace())
                     if cfg_dest.get('class_path') == val:
                         return
-                    elif cfg_dest.get('init_args') is not None and cfg_dest.get('init_args') != {}:
+                    elif cfg_dest.get('init_args') is not None and not is_empty_namespace(cfg_dest.get('init_args')):
                         warnings.warn(
-                            'Argument ' + opt_str + '=' + val + ' implies discarding init_args ' + str(cfg_dest.get('init_args')) +
-                            ' defined for class_path ' + cfg_dest.get('class_path')
+                            f'Argument {opt_str}={val} implies discarding init_args {cfg_dest.get("init_args").as_dict()} '
+                            f'defined for class_path {cfg_dest.get("class_path")}'
                         )
-                    val = {'class_path': val}
+                    val = Namespace(class_path=val)
                 elif '.init_args.' in opt_str:
                     match = re.match(r'.+\.init_args\.([^.]+)$', opt_str)
-                    if match and isinstance(getattr(cfg, self.dest, None), dict):
-                        init_args = getattr(cfg, self.dest).get('init_args', {})
-                        init_args[match.groups()[0]] = val
-                        val = getattr(cfg, self.dest)
-                    else:
-                        cfg_dest = get_key_value_from_flat_dict(vars(cfg), self.dest)
-                        init_args = cfg_dest.get('init_args', {})
-                        init_args[match.groups()[0]] = val
-                        val = cfg_dest
+                    cfg_dest = cfg.get(self.dest, Namespace())
+                    init_args = cfg_dest.get('init_args', Namespace())
+                    init_args[match.groups()[0]] = val
+                    val = cfg_dest
             val = self._check_type(val)
         if 'cfg_dest' in locals():
-            update_key_value_in_flat_dict(vars(args[1]), self.dest, val)
+            args[1].update(val, self.dest)
         else:
             setattr(args[1], self.dest, val)
 
@@ -239,7 +277,7 @@ class ActionTypeHint(Action):
             try:
                 orig_val = val
                 try:
-                    val, config_path = _load_config(val, enable_path=self._enable_path, flat_namespace=False)
+                    val, config_path = _load_config(val, enable_path=self._enable_path)
                 except (yamlParserError, yamlScannerError):
                     config_path = None
                 path_meta = val.pop('__path__') if isinstance(val, dict) and '__path__' in val else None
@@ -252,25 +290,47 @@ class ActionTypeHint(Action):
                         val = adapt_typehints(orig_val, self._typehint, sub_add_kwargs=sub_add_kwargs)
                     else:
                         if self._enable_path and config_path is None and isinstance(orig_val, str):
-                            msg = '\n- Expected a config path but "'+orig_val+'" either not accessible or invalid.\n- '
+                            msg = f'\n- Expected a config path but "{orig_val}" either not accessible or invalid.\n- '
                             raise type(ex)(msg+str(ex)) from ex
                         raise ex
                 if path_meta is not None:
                     val['__path__'] = path_meta
-                if isinstance(val, dict) and config_path is not None:
+                if isinstance(val, (Namespace, dict)) and config_path is not None:
                     val['__path__'] = config_path
                 value[num] = val
             except (TypeError, ValueError) as ex:
-                elem = '' if not islist else ' element '+str(num+1)
-                raise TypeError('Parser key "'+self.dest+'"'+elem+': '+str(ex)) from ex
+                elem = '' if not islist else f' element {num+1}'
+                raise TypeError(f'Parser key "{self.dest}"{elem}: {ex}') from ex
         return value if islist else value[0]
 
 
-    def _instantiate_classes(self, val):
+    def instantiate_classes(self, val):
         sub_add_kwargs = getattr(self, 'sub_add_kwargs', {})
         return adapt_typehints(val, self._typehint, instantiate_classes=True, sub_add_kwargs=sub_add_kwargs)
 
 
+    def get_class_and_init_args(self, value):
+        final_class = ActionTypeHint.is_final_class_typehint(self)
+        if final_class:
+            typehint_origin = getattr(self._typehint, '__origin__', None)
+            class_type = [a for a in self._typehint.__args__ if a != NoneType][0] if typehint_origin == Union else self._typehint
+            init_args = value
+        else:
+            class_type = import_object(value['class_path'])
+            init_args = value.get('init_args')
+        return class_type, init_args, final_class
+
+
+    @staticmethod
+    def get_class_parser(val_class, sub_add_kwargs=None):
+        from .core import ArgumentParser
+        if isinstance(val_class, str):
+            val_class = import_object(val_class)
+        parser = ArgumentParser(error_handler=None)
+        parser.add_class_arguments(val_class, **(sub_add_kwargs or {}))
+        return parser
+
+
     def completer(self, prefix, **kwargs):
         """Used by argcomplete, validates value and shows expected type."""
         if self._typehint == bool:
@@ -314,19 +374,19 @@ def adapt_typehints(val, typehint, serialize=False, instantiate_classes=False, s
             val = adapt_typehints(val, type_val, **adapt_kwargs)
         elif isinstance(val, str):
             try:
-                val = _load_config(val, enable_path=False, flat_namespace=False)[0]
+                val, _ = _load_config(val, enable_path=False)
             except (yamlParserError, yamlScannerError):
                 pass
 
     # Literal
     elif typehint_origin == Literal:
         if val not in subtypehints:
-            raise ValueError('Expected a '+str(typehint)+' but got "'+str(val)+'"')
+            raise ValueError(f'Expected a {typehint} but got "{val}"')
 
     # Basic types
     elif typehint in leaf_types:
         if not isinstance(val, typehint):
-            raise ValueError('Expected a '+str(typehint)+' but got "'+str(val)+'"')
+            raise ValueError(f'Expected a {typehint} but got "{val}"')
 
     # Registered types
     elif typehint in registered_types:
@@ -345,7 +405,8 @@ def adapt_typehints(val, typehint, serialize=False, instantiate_classes=False, s
             if isinstance(val, typehint):
                 val = val.name
             else:
-                typehint[val]
+                if val not in typehint.__members__:
+                    raise ValueError(f'Value "{val}" is not a valid member name for the enum "{typehint}"')
         elif not serialize and not isinstance(val, typehint):
             val = typehint[val]
 
@@ -358,7 +419,7 @@ def adapt_typehints(val, typehint, serialize=False, instantiate_classes=False, s
             val = import_object(val)
             if (typehint in {Type, type} and not isinstance(val, type)) or \
                (typehint not in {Type, type} and not _issubclass(val, subtypehints[0])):
-                raise ValueError('Value "'+str(path)+'" is not a '+str(typehint)+'.')
+                raise ValueError(f'Value "{path}" is not a {typehint}.')
 
     # Union
     elif typehint_origin == Union:
@@ -371,19 +432,19 @@ def adapt_typehints(val, typehint, serialize=False, instantiate_classes=False, s
                 vals.append(ex)
         if all(isinstance(v, Exception) for v in vals):
             e = indent_text('\n- '.join(str(v) for v in ['']+vals))
-            raise ValueError('Value "'+str(val)+'" does not validate against any of the types in '+str(typehint)+':'+e)
+            raise ValueError(f'Value "{val}" does not validate against any of the types in {typehint}:{e}')
         val = [v for v in vals if not isinstance(v, Exception)][0]
 
     # Tuple or Set
-    elif typehint_origin in {Tuple, tuple, Set, set, frozenset, MutableSet, abcSet, abcMutableSet}:
+    elif typehint_origin in tuple_set_origin_types:
         if not isinstance(val, (list, tuple, set)):
-            raise ValueError('Expected a '+str(typehint_origin)+' but got "'+str(val)+'"')
+            raise ValueError(f'Expected a {typehint_origin} but got "{val}"')
         val = list(val)
         if subtypehints is not None:
             is_tuple = typehint_origin in {Tuple, tuple}
             is_ellipsis = is_ellipsis_tuple(typehint)
             if is_tuple and not is_ellipsis and len(val) != len(subtypehints):
-                raise ValueError('Expected a tuple with '+str(len(subtypehints))+' elements but got "'+str(val)+'"')
+                raise ValueError(f'Expected a tuple with {len(subtypehints)} elements but got "{val}"')
             for n, v in enumerate(val):
                 subtypehint = subtypehints[0 if is_ellipsis else n]
                 val[n] = adapt_typehints(v, subtypehint, **adapt_kwargs)
@@ -393,82 +454,110 @@ def adapt_typehints(val, typehint, serialize=False, instantiate_classes=False, s
             val = tuple(val) if typehint_origin in {Tuple, tuple} else set(val)
 
     # List, Iterable or Sequence
-    elif typehint_origin in {List, list, Iterable, Sequence, MutableSequence, abcIterable, abcSequence,
-                             abcMutableSequence}:
+    elif typehint_origin in sequence_origin_types:
         if not isinstance(val, list):
-            raise ValueError('Expected a List but got "'+str(val)+'"')
+            raise ValueError(f'Expected a List but got "{val}"')
         if subtypehints is not None:
             for n, v in enumerate(val):
                 val[n] = adapt_typehints(v, subtypehints[0], **adapt_kwargs)
 
-    # Dict
-    elif typehint_origin in {Dict, dict, Mapping, MutableMapping, abcMapping, abcMutableMapping}:
+    # Dict, Mapping
+    elif typehint_origin in mapping_origin_types:
         if not isinstance(val, dict):
-            raise ValueError('Expected a Dict but got "'+str(val)+'"')
+            raise ValueError(f'Expected a Dict but got "{val}"')
         if subtypehints is not None:
             if subtypehints[0] == int:
                 cast = str if serialize else int
                 val = {cast(k): v for k, v in val.items()}
             for k, v in val.items():
-                val[k] = adapt_typehints(v, subtypehints[1], **adapt_kwargs)
+                if "linked_targets" in adapt_kwargs["sub_add_kwargs"]:
+                    kwargs = copy.deepcopy(adapt_kwargs)
+                    sub_add_kwargs = kwargs["sub_add_kwargs"]
+                    sub_add_kwargs["linked_targets"] = {t[len(k + "."):] for t in sub_add_kwargs["linked_targets"]
+                                                        if t.startswith(k + ".")}
+                    sub_add_kwargs["linked_targets"] = {t[len("init_args."):] if t.startswith("init_args.") else t
+                                                        for t in sub_add_kwargs["linked_targets"]}
+                else:
+                    kwargs = adapt_kwargs
+                val[k] = adapt_typehints(v, subtypehints[1], **kwargs)
 
     # Final class
     elif is_final_class(typehint):
-        if not isinstance(val, dict):
-            raise ValueError('Expected a Dict but got "'+str(val)+'"')
+        if isinstance(val, dict):
+            val = Namespace(val)
+        if not isinstance(val, Namespace):
+            raise ValueError(f'Expected a Dict/Namespace but got "{val}"')
         val = adapt_class_type(typehint, val, serialize, instantiate_classes, sub_add_kwargs)
 
     # Subclass
     elif not hasattr(typehint, '__origin__') and inspect.isclass(typehint):
         if isinstance(val, typehint):
-            if isinstance(val, LazyInitBaseClass):
-                val = val.get_init_dict()
-            elif serialize:
+            if serialize:
                 val = str(val)
-                warnings.warn(
-                    'Not possible to serialize an instance of ' + str(typehint.__name__) + '. It will be represented as the '
-                    'string ' + val + '. If this was set as a default, consider setting a dict or using lazy_instance.'
-                )
+                warning(f"""
+                    Not possible to serialize an instance of {typehint}. It will
+                    be represented as the string {val}. If this was set as a
+                    default, consider using lazy_instance.
+                """)
             return val
         if serialize and isinstance(val, str):
             return val
-        if not (isinstance(val, str) or (isinstance(val, dict) and 'class_path' in val) or (isinstance(val, Namespace) and hasattr(val, 'class_path'))):
-            raise ValueError('Type '+str(typehint)+' expects an str or a Dict with a class_path entry but got "'+str(val)+'"')
+        if not (isinstance(val, str) or is_class_object(val)):
+            raise ValueError(f'Type {typehint} expects an str or a Dict/Namespace with a class_path entry but got "{val}"')
         try:
-            if isinstance(val, Namespace):
-                val = namespace_to_dict(val)
             if isinstance(val, str):
-                val_class = import_object(val)
-                val = {'class_path': val}
-            else:
-                val_class = import_object(val['class_path'])
+                val = Namespace(class_path=val)
+            elif isinstance(val, dict):
+                val = Namespace(val)
+            val_class = import_object(val['class_path'])
+            if isinstance(val.get('init_args'), dict):
+                val['init_args'] = Namespace(val['init_args'])
             if not _issubclass(val_class, typehint):
-                raise ValueError('"'+val['class_path']+'" is not a subclass of '+typehint.__name__)
-            init_args = val.get('init_args', {})
+                raise ValueError(f'"{val["class_path"]}" is not a subclass of {typehint.__name__}')
+            init_args = val.get('init_args', Namespace())
             adapted = adapt_class_type(val_class, init_args, serialize, instantiate_classes, sub_add_kwargs)
             if instantiate_classes and sub_add_kwargs.get('instantiate', True):
                 val = adapted
-            else:
+            elif adapted is not None and not is_empty_namespace(adapted):
                 val['init_args'] = adapted
-        except (ImportError, ModuleNotFound, AttributeError, AssertionError, ParserError) as ex:
+        except (ImportError, AttributeError, AssertionError, ParserError) as ex:
             class_path = val if isinstance(val, str) else val['class_path']
-            e = indent_text('\n- '+str(ex))
-            raise ValueError('Problem with given class_path "'+class_path+'":'+e) from ex
+            e = indent_text(f'\n- {ex}')
+            raise ValueError(f'Problem with given class_path "{class_path}":{e}') from ex
 
     return val
 
 
+def is_class_object(val):
+    return isinstance(val, (dict, Namespace)) and 'class_path' in val
+
+
 def adapt_class_type(val_class, init_args, serialize, instantiate_classes, sub_add_kwargs):
-    from .core import ArgumentParser
-    parser = ArgumentParser(error_handler=None, parse_as_dict=True)
-    parser.add_class_arguments(val_class, **sub_add_kwargs)
+    parser = ActionTypeHint.get_class_parser(val_class, sub_add_kwargs)
+
+    # No need to re-create the linked arg but just "inform" the corresponding parser actions that it exists upstream.
+    for target in sub_add_kwargs.get('linked_targets', []):
+        split_index = target.find(".")
+        if split_index != -1:
+            split = ".init_args." if target[split_index:].startswith(".init_args.") else "."
+
+            parent_key, key = target.split(split, maxsplit=1)
+
+            action = next(a for a in parser._actions if a.dest == parent_key)
+
+            sub_add_kwargs = getattr(action, 'sub_add_kwargs')
+            sub_add_kwargs.setdefault('linked_targets', set())
+            sub_add_kwargs['linked_targets'].add(key)
+
+            break
+
     if instantiate_classes:
-        init_args = parser.instantiate_subclasses(init_args)
+        init_args = parser.instantiate_classes(init_args)
         if not sub_add_kwargs.get('instantiate', True):
             return init_args
         return val_class(**init_args)
     if serialize:
-        init_args = yaml.safe_load(parser.dump(init_args))
+        init_args = None if is_empty_namespace(init_args) else yaml.safe_load(parser.dump(init_args))
     else:
         init_args = parser.parse_object(init_args)
     return init_args
@@ -487,6 +576,12 @@ def is_optional(annotation, ref_type):
         any(_issubclass(a, ref_type) for a in annotation.__args__)
 
 
+def typehint_from_action(action_or_typehint):
+    if isinstance(action_or_typehint, Action):
+        action_or_typehint = getattr(action_or_typehint, '_typehint', None)
+    return action_or_typehint
+
+
 def type_to_str(obj):
     if _issubclass(obj, (bool, int, float, str, Enum)):
         return obj.__name__
@@ -517,26 +612,34 @@ class LazyInitBaseClass:
         self._lazy_class_path = get_import_path(class_type)
         self._lazy_kwargs = lazy_kwargs
         self._lazy_methods = {}
-        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
-            if name in {'__init__', '_lazy_init', '_lazy_init_then_call_method', 'get_init_dict'}:
+        seen_methods: Dict = {}
+        for name, _ in inspect.getmembers(class_type, predicate=inspect.isfunction):
+            method = getattr(self, name)
+            if not inspect.ismethod(method) or \
+               name in {'__init__', '_lazy_init', '_lazy_init_then_call_method', 'lazy_get_init_data'}:
                 continue
+            assert name not in self.__dict__
             self._lazy_methods[name] = method
-            setattr(self, name, partial(self._lazy_init_then_call_method, method_name=name))
+            if method in seen_methods:
+                self.__dict__[name] = seen_methods[method]
+            else:
+                self.__dict__[name] = partial(self._lazy_init_then_call_method, name)
+                seen_methods[method] = self.__dict__[name]
 
     def _lazy_init(self):
-        for name, method in self._lazy_methods.items():
-            setattr(self, name, method)
+        for name in self._lazy_methods.keys():
+            del self.__dict__[name]
         super().__init__(**self._lazy_kwargs)
 
     def _lazy_init_then_call_method(self, method_name, *args, **kwargs):
         self._lazy_init()
         return getattr(self, method_name)(*args, **kwargs)
 
-    def get_init_dict(self):
-        init_dict = {'class_path': self._lazy_class_path}
+    def lazy_get_init_data(self):
+        init = Namespace(class_path=self._lazy_class_path)
         if len(self._lazy_kwargs) > 0:
-            init_dict['init_args'] = self._lazy_kwargs
-        return init_dict
+            init['init_args'] = Namespace(self._lazy_kwargs)
+        return init
 
 
 ClassType = TypeVar('ClassType')
@@ -545,17 +648,17 @@ ClassType = TypeVar('ClassType')
 def lazy_instance(class_type: Type[ClassType], **kwargs) -> ClassType:
     """Instantiates a lazy instance of the given type.
 
-    By lazy it is meant that the init is delayed unit the first time that a
-    method of the instance is called. It also provides a `get_init_dict` method
+    By lazy it is meant that the __init__ is delayed unit the first time that a
+    method of the instance is called. It also provides a `lazy_get_init_data` method
     useful for serializing.
 
-    Note: Only supported in CPython.
-
     Args:
         class_type: The class to instantiate.
         **kwargs: Any keyword arguments to use for instantiation.
     """
-    class LazyInitClass(LazyInitBaseClass, class_type):  # type: ignore
-        pass
-
-    return LazyInitClass(class_type, kwargs)
+    lazy_init_class = type(
+        'LazyInstance_'+class_type.__name__,
+        (LazyInitBaseClass, class_type),
+        {'__doc__': f'Class for lazy instances of {class_type}'},
+    )
+    return lazy_init_class(class_type, kwargs)
