@@ -3,14 +3,14 @@
 import os
 import sys
 import unittest
+import warnings
 
 
 testing_package = os.path.basename(os.path.dirname(os.path.realpath(__file__)))
 
 
 def run_tests():
-    pattern = '*_tests.py' if sys.version_info[:2] < (3, 6) else '*_tests*.py'
-    tests = unittest.defaultTestLoader.discover(testing_package, pattern=pattern)
+    tests = unittest.defaultTestLoader.discover(testing_package, pattern='*_tests.py')
     if not unittest.TextTestRunner(verbosity=2).run(tests).wasSuccessful():
         sys.exit(True)
 
@@ -21,6 +21,8 @@ def run_test_coverage():
     except:
         print('error: coverage package not found, run_test_coverage requires it.')
         sys.exit(True)
+    warnings.simplefilter('default')
+    os.environ['PYTHONWARNINGS'] = 'default'
     package_source = os.path.dirname(__file__.replace('_tests', ''))
     cov = coverage.Coverage(source=[package_source])
     cov.start()
