@@ -92,22 +92,24 @@ class SignaturesTests(unittest.TestCase):
             self.assertIsNone(_find_action(parser, key), key+' should not be in parser but is')
 
         cfg = parser.parse_args(['--c3_a0=0', '--c3_a3=true', '--c3_a4=a'], with_meta=False)
-        self.assertEqual(namespace_to_dict(cfg), {'c1_a2': 2.0,
-                                                  'c1_a4': 4,
-                                                  'c1_a5': 'five',
-                                                  'c3_a0': 0,
-                                                  'c3_a1': '1',
-                                                  'c3_a2': 2.0,
-                                                  'c3_a3': True,
-                                                  'c3_a4': 'a',
-                                                  'c3_a5': 5,
-                                                  'c3_a6': None,
-                                                  'c3_a7': ('7', 7, 7.0),
-                                                  'c3_a8': None})
+        self.assertEqual(cfg.as_dict(), {
+            'c1_a2': 2.0,
+            'c1_a4': 4,
+            'c1_a5': 'five',
+            'c3_a0': 0,
+            'c3_a1': '1',
+            'c3_a2': 2.0,
+            'c3_a3': True,
+            'c3_a4': 'a',
+            'c3_a5': 5,
+            'c3_a6': None,
+            'c3_a7': ('7', 7, 7.0),
+            'c3_a8': None,
+        })
         self.assertEqual([1, 2], parser.parse_args(['--c3_a0=0', '--c3_a5=[1,2]']).c3_a5)
-        self.assertEqual({'k': 5.0}, namespace_to_dict(parser.parse_args(['--c3_a0=0', '--c3_a5={"k": 5.0}']).c3_a5))
+        self.assertEqual({'k': 5.0}, parser.parse_args(['--c3_a0=0', '--c3_a5={"k": 5.0}']).c3_a5)
         self.assertEqual(('3', 3, 3.0), parser.parse_args(['--c3_a0=0', '--c3_a7=["3", 3, 3.0]']).c3_a7)
-        self.assertEqual('a', Class3(**namespace_to_dict(cfg))())
+        self.assertEqual('a', Class3(**cfg.as_dict())())
 
         self.assertRaises(ParserError, lambda: parser.parse_args([]))  # c3_a0 is required
         self.assertRaises(ParserError, lambda: parser.parse_args(['--c3_a0=0', '--c3_a7=["3", "3", 3.0]']))  # tuple[1] is int
@@ -150,7 +152,7 @@ class SignaturesTests(unittest.TestCase):
         def func(a1: Union[int, Dict[int, int]] = 1):
             pass
 
-        parser = ArgumentParser()
+        parser = ArgumentParser(error_handler=None)
         parser.add_function_arguments(func)
         parser.get_defaults()
         cfg = parser.parse_args(['--a1={"2": 7, "4": 9}'])
@@ -213,7 +215,7 @@ class SignaturesTests(unittest.TestCase):
         for key in ['m.a4', 's.a3']:
             self.assertIsNone(_find_action(parser, key), key+' should not be in parser but is')
 
-        cfg = namespace_to_dict(parser.parse_args(['--m.a1=x', '--s.a1=y'], with_meta=False))
+        cfg = parser.parse_args(['--m.a1=x', '--s.a1=y'], with_meta=False).as_dict()
         self.assertEqual(cfg, {'m': {'a1': 'x', 'a2': 2.0, 'a3': False}, 's': {'a1': 'y', 'a2': 2.0}})
         self.assertEqual('x', MyClass().mymethod(**cfg['m']))
         self.assertEqual('y', MyClass.mystaticmethod(**cfg['s']))
@@ -253,7 +255,7 @@ class SignaturesTests(unittest.TestCase):
             self.assertIsNotNone(_find_action(parser, key), key+' should be in parser but is not')
         self.assertIsNone(_find_action(parser, 'a4'), 'a4 should not be in parser but is')
 
-        cfg = namespace_to_dict(parser.parse_args(['--a1=x'], with_meta=False))
+        cfg = parser.parse_args(['--a1=x'], with_meta=False).as_dict()
         self.assertEqual(cfg, {'a1': 'x', 'a2': 2.0, 'a3': False})
         self.assertEqual('x', func(**cfg))
 
@@ -264,24 +266,24 @@ class SignaturesTests(unittest.TestCase):
 
 
     def test_add_subclass_arguments(self):
-        parser = ArgumentParser(error_handler=None, parse_as_dict=True)
+        parser = ArgumentParser(error_handler=None)
         parser.add_subclass_arguments(calendar.Calendar, 'cal')
 
         cal = {'class_path': 'calendar.Calendar', 'init_args': {'firstweekday': 1}}
         cfg = parser.parse_args(['--cal='+json.dumps(cal)])
-        self.assertEqual(cfg['cal'], cal)
+        self.assertEqual(cfg['cal'].as_dict(), cal)
 
         cal['init_args']['firstweekday'] = 2
         cfg = parser.parse_args(['--cal.class_path=calendar.Calendar', '--cal.init_args.firstweekday=2'])
-        self.assertEqual(cfg['cal'], cal)
+        self.assertEqual(cfg['cal'].as_dict(), cal)
 
         cal['init_args']['firstweekday'] = 3
         cfg = parser.parse_args(['--cal.class_path', 'calendar.Calendar', '--cal.init_args.firstweekday', '3'])
-        self.assertEqual(cfg['cal'], cal)
+        self.assertEqual(cfg['cal'].as_dict(), cal)
 
         cal['init_args']['firstweekday'] = 4
         cfg = parser.parse_args(['--cal.class_path=calendar.Calendar', '--cal.init_args.firstweekday=4', '--cal.class_path=calendar.Calendar'])
-        self.assertEqual(cfg['cal'], cal)
+        self.assertEqual(cfg['cal'].as_dict(), cal)
 
         self.assertRaises(ParserError, lambda: parser.parse_args(['--cal={"class_path":"not.exist.Class"}']))
         self.assertRaises(ParserError, lambda: parser.parse_args(['--cal={"class_path":"calendar.January"}']))
@@ -293,12 +295,24 @@ class SignaturesTests(unittest.TestCase):
             parser.parse_args(['--cal.help=calendar.Calendar'])
         self.assertIn('--cal.init_args.firstweekday', out.getvalue())
 
-        if platform.python_implementation() == 'CPython':
+        if True:  # platform.python_implementation() == 'CPython':
+            class MyCalendar(calendar.Calendar):
+                init_called = False
+                getfirst = calendar.Calendar.getfirstweekday
+                def __init__(self, *args, **kwargs):
+                    self.init_called = True
+                    super().__init__(*args, **kwargs)
+
+            lazy_calendar = lazy_instance(MyCalendar, firstweekday=3)
+            self.assertFalse(lazy_calendar.init_called, '__init__ was already called but supposed to be lazy')
+            self.assertEqual(lazy_calendar.getfirstweekday(), 3)
+            self.assertTrue(lazy_calendar.init_called)
+
             cal['init_args']['firstweekday'] = 4
             lazy_calendar = lazy_instance(calendar.Calendar, firstweekday=4)
             parser.set_defaults({'cal': lazy_calendar})
             cfg = parser.parse_string(parser.dump(parser.parse_args([])))
-            self.assertEqual(cfg['cal'], cal)
+            self.assertEqual(cfg['cal'].as_dict(), cal)
             self.assertEqual(lazy_calendar.getfirstweekday(), 4)
 
             parser.add_argument('--config', action=ActionConfigFile)
@@ -307,17 +321,21 @@ class SignaturesTests(unittest.TestCase):
                 parser.parse_args(['--print_config'])
             self.assertIn('class_path: calendar.Calendar', out.getvalue())
 
+            out = StringIO()
+            parser.print_help(out)
+            self.assertIn("'init_args': {'firstweekday': 4}", out.getvalue())
+
         parser.set_defaults({'cal': calendar.Calendar(firstweekday=4)})
         cfg = parser.parse_args([])
         self.assertIsInstance(cfg['cal'], calendar.Calendar)
-        cfg = parser.instantiate_subclasses(cfg)
+        cfg = parser.instantiate_classes(cfg)
         self.assertIsInstance(cfg['cal'], calendar.Calendar)
         dump = parser.dump(cfg)
         self.assertIn('cal: <calendar.Calendar object at ', dump)
 
 
     def test_add_subclass_discard_init_args(self):
-        parser = ArgumentParser(error_handler=None, parse_as_dict=True)
+        parser = ArgumentParser(error_handler=None)
         parser.add_subclass_arguments(calendar.Calendar, 'cal')
 
         class MyCal(calendar.Calendar):
@@ -383,10 +401,12 @@ class SignaturesTests(unittest.TestCase):
 
 
     def test_not_required_group(self):
-        parser = ArgumentParser(parse_as_dict=True, error_handler=None)
+        parser = ArgumentParser(error_handler=None)
         parser.add_subclass_arguments(calendar.Calendar, 'cal', required=False)
         cfg = parser.parse_args([])
-        self.assertEqual(cfg, {'cal': {}})
+        self.assertEqual(cfg, Namespace())
+        cfg_init = parser.instantiate_classes(cfg)
+        self.assertEqual(cfg_init, Namespace())
 
 
     def test_invalid_type(self):
@@ -439,7 +459,7 @@ class SignaturesTests(unittest.TestCase):
         def func(a1: Any = MyEnum.B):
             return a1
 
-        parser = ArgumentParser(error_handler=None, parse_as_dict=True)
+        parser = ArgumentParser(error_handler=None)
         parser.add_function_arguments(func)
         cfg = parser.parse_args([])
         self.assertEqual('a1: B\n', parser.dump(cfg))
@@ -530,6 +550,7 @@ class SignaturesTests(unittest.TestCase):
         parser.add_class_arguments(ClassB, 'b')
         cfg = parser.parse_args(['--b.b2={"a2": 6.7}'])
         self.assertEqual(cfg.b.b2, Namespace(a1=1, a2=6.7))
+        self.assertEqual(cfg, parser.parse_string(parser.dump(cfg)))
         cfg = parser.instantiate_classes(cfg)
         self.assertIsInstance(cfg['b'], ClassB)
         self.assertIsInstance(cfg['b'].b2, ClassA)
@@ -559,6 +580,19 @@ class SignaturesTests(unittest.TestCase):
         self.assertRaises(ParserError, lambda: parser.parse_args(['--a2=-1']))
 
 
+    def test_dict_int_str_type(self):
+        class Foo:
+            def __init__(self, d: Dict[int, str]):
+                self.d = d
+
+        parser = ArgumentParser(error_handler=None)
+        parser.add_class_arguments(Foo)
+        parser.add_argument('--config', action=ActionConfigFile)
+        cfg = {'d': {1: 'val1', 2: 'val2'}}
+        self.assertEqual(cfg['d'], parser.parse_args(['--config', str(cfg)]).d)
+        self.assertRaises(ParserError, lambda: parser.parse_args(['--config={"d": {"a": "b"}}']))
+
+
     def test_logger_debug(self):
 
         with _suppress_stderr():
@@ -628,7 +662,7 @@ class SignaturesTests(unittest.TestCase):
         cfg = parser.parse_args(['--c1={'+class_path+', '+init_args+'}'])
         self.assertEqual(cfg.c1.class_path, 'jsonargparse_tests.signatures_tests.Class1')
         self.assertEqual(cfg.c1.init_args, Namespace(a1=7, a2=2.3))
-        cfg = parser.instantiate_subclasses(cfg)
+        cfg = parser.instantiate_classes(cfg)
         self.assertIsInstance(cfg['c1'], Class1)
         self.assertEqual(7, cfg['c1'].a1)
         self.assertEqual(2.3, cfg['c1'].a2)
@@ -656,7 +690,7 @@ class SignaturesTests(unittest.TestCase):
             def __init__(self, a: int = 1):
                 self.a = a
 
-        parser = ArgumentParser(parse_as_dict=True)
+        parser = ArgumentParser()
         subcommands = parser.add_subcommands()
         subparser = ArgumentParser()
         key = "foo"
@@ -764,6 +798,9 @@ class SignaturesTests(unittest.TestCase):
         self.assertEqual(11+7, cfg.b.v2)
         self.assertEqual(Namespace(), parser.parse_args([], defaults=False))
 
+        dump = yaml.safe_load(parser.dump(cfg))
+        self.assertEqual(dump, {'a': {'v1': 11, 'v2': 7}, 'b': {'v3': 2}})
+
         self.assertRaises(ParserError, lambda: parser.parse_args(['--b.v1=5']))
         self.assertRaises(ValueError, lambda: parser.link_arguments('a.v2', 'b.v1'))
         self.assertRaises(ValueError, lambda: parser.link_arguments('x', 'b.v2'))
@@ -809,6 +846,14 @@ class SignaturesTests(unittest.TestCase):
         self.assertEqual(cfg.c.init_args.firstweekday, 4)
         self.assertEqual(cfg.c.init_args.firstweekday, cfg.a.init_args.v1+cfg.a.init_args.v2)
 
+        cfg_init = parser.instantiate_classes(cfg)
+        self.assertIsInstance(cfg_init.a, ClassA)
+        self.assertIsInstance(cfg_init.c, calendar.Calendar)
+
+        dump = yaml.safe_load(parser.dump(cfg))
+        a_value['init_args']['v1'] = 1
+        self.assertEqual(dump, {'a': a_value, 'c': {'class_path': 'calendar.Calendar'}})
+
         self.assertRaises(ValueError, lambda: parser.link_arguments('a.init_args.v1', 'c'))
 
         a_value['init_args'] = {'v1': 'a', 'v2': 'b'}
@@ -816,12 +861,87 @@ class SignaturesTests(unittest.TestCase):
             parser.parse_args(['--a='+json.dumps(a_value), '--c=calendar.Calendar'])
 
 
+    def test_link_arguments_subclasses_with_instantiate_false(self):
+        class ClassA:
+            def __init__(
+                self,
+                v: Union[int, str] = 1,
+                c: Optional[calendar.Calendar] = None,
+            ):
+                self.c = c
+
+        from jsonargparse_tests import signatures_tests
+        setattr(signatures_tests, 'ClassA', ClassA)
+
+        parser = ArgumentParser(error_handler=None)
+        parser.add_subclass_arguments(ClassA, 'a')
+        parser.add_subclass_arguments(calendar.Calendar, 'c', instantiate=False)
+        parser.link_arguments('c', 'a.init_args.c')
+
+        a_value = {'class_path': 'jsonargparse_tests.signatures_tests.ClassA'}
+        c_value = {
+            'class_path': 'calendar.Calendar',
+            'init_args': {
+                'firstweekday': 3,
+            },
+        }
+
+        cfg = parser.parse_args(['--a='+json.dumps(a_value), '--c='+json.dumps(c_value)])
+        self.assertEqual(cfg.c.as_dict(), {'class_path': 'calendar.Calendar', 'init_args': {'firstweekday': 3}})
+        self.assertEqual(cfg.c, cfg.a.init_args.c)
+
+        cfg_init = parser.instantiate_classes(cfg)
+        self.assertIsInstance(cfg_init.c, Namespace)
+        self.assertIsInstance(cfg_init.a, ClassA)
+        self.assertIsInstance(cfg_init.a.c, calendar.Calendar)
+        self.assertEqual(cfg_init.a.c.firstweekday, 3)
+
+        dump = yaml.safe_load(parser.dump(cfg))
+        self.assertNotIn('c', dump['a']['init_args'])
+
+
+    def test_link_arguments_subclass_as_dict(self):
+        class ClassA:
+            def __init__(
+                self,
+                a1: dict,
+                a2: Optional[dict] = None,
+                a3: Any = None,
+            ):
+                pass
+
+        def return_dict(value: dict):
+            return value
+
+        from jsonargparse_tests import signatures_tests
+        setattr(signatures_tests, 'ClassA', ClassA)
+
+        parser = ArgumentParser(error_handler=None)
+        parser.add_subclass_arguments(ClassA, 'a')
+        parser.add_subclass_arguments(calendar.Calendar, 'c')
+        parser.link_arguments('c', 'a.init_args.a1', compute_fn=return_dict)
+        parser.link_arguments('c', 'a.init_args.a2')
+        parser.link_arguments('c', 'a.init_args.a3')
+
+        a_value = {'class_path': 'jsonargparse_tests.signatures_tests.ClassA'}
+        c_value = {
+            'class_path': 'calendar.Calendar',
+            'init_args': {
+                'firstweekday': 3,
+            },
+        }
+
+        cfg = parser.parse_args(['--a='+json.dumps(a_value), '--c='+json.dumps(c_value)])
+        self.assertEqual(cfg.a.init_args.a1, c_value)
+        self.assertEqual(cfg.a.init_args.a2, c_value)
+
+
     def test_link_arguments_subcommand(self):
         class Foo:
             def __init__(self, a: int):
                 self.a = a
 
-        parser = ArgumentParser(parse_as_dict=True)
+        parser = ArgumentParser()
         subparser = ArgumentParser()
 
         subcommands = parser.add_subcommands()
@@ -831,7 +951,7 @@ class SignaturesTests(unittest.TestCase):
         subcommands.add_subcommand('cmd', subparser)
 
         cfg = parser.parse_args(['cmd', '--b=2'])
-        self.assertEqual(cfg['cmd']['foo'], {'a': 2})
+        self.assertEqual(cfg['cmd']['foo'].as_dict(), {'a': 2})
 
         cfg = parser.instantiate_classes(cfg)
         self.assertIsInstance(cfg['cmd']['foo'], Foo)
@@ -1012,11 +1132,11 @@ class SignaturesConfigTests(TempDirTestCase):
 
         cfg = parser.parse_args(['--cfg', cfg_path])
         self.assertEqual(str(cfg.func.__path__), subcfg_path)
-        self.assertEqual(strip_meta(cfg.func), {'a1': 'one', 'a2': 2.0, 'a3': True})
+        self.assertEqual(strip_meta(cfg.func), Namespace(a1='one', a2=2.0, a3=True))
 
 
     def test_add_subclass_arguments_with_config(self):
-        parser = ArgumentParser(error_handler=None, parse_as_dict=True)
+        parser = ArgumentParser(error_handler=None)
         parser.add_argument('--cfg', action=ActionConfigFile)
         parser.add_subclass_arguments(calendar.Calendar, 'cal')
 
@@ -1026,11 +1146,11 @@ class SignaturesConfigTests(TempDirTestCase):
             f.write(yaml.dump({'cal': cal}))
 
         cfg = parser.parse_args(['--cfg='+cfg_path])
-        self.assertEqual(cfg['cal'], cal)
+        self.assertEqual(cfg['cal'].as_dict(), cal)
 
         cal['init_args']['firstweekday'] = 2
         cfg = parser.parse_args(['--cfg='+cfg_path, '--cal.init_args.firstweekday=2'])
-        self.assertEqual(cfg['cal'], cal)
+        self.assertEqual(cfg['cal'].as_dict(), cal)
 
 
     def test_add_class_arguments_with_config_not_found(self):
@@ -1068,5 +1188,216 @@ class SignaturesConfigTests(TempDirTestCase):
             self.assertEqual({'class_path': 'calendar.Calendar', 'init_args': {'firstweekday': 0}}, cal)
 
 
+@unittest.skipIf(not dataclasses_support, 'dataclasses package is required')
+class DataclassesTests(unittest.TestCase):
+
+    @classmethod
+    def setUpClass(cls):
+        cls.dataclasses = import_dataclasses('SignaturesTests')
+
+        @cls.dataclasses.dataclass
+        class MyDataClassA:
+            """MyDataClassA description
+
+            Args:
+                a1: a1 help
+                a2: a2 help
+            """
+            a1: PositiveInt = PositiveInt(1)
+            a2: str = '2'
+
+        @cls.dataclasses.dataclass
+        class MyDataClassB:
+            """MyDataClassB description
+
+            Args:
+                b1: b1 help
+                b2: b2 help
+            """
+            b1: PositiveFloat = PositiveFloat(3.0)
+            b2: MyDataClassA = MyDataClassA()
+
+        cls.MyDataClassA = MyDataClassA
+        cls.MyDataClassB = MyDataClassB
+
+
+    @unittest.skipIf(not docstring_parser_support, 'docstring-parser package is required')
+    def test_add_dataclass_arguments(self):
+        parser = ArgumentParser(error_handler=None)
+        parser.add_dataclass_arguments(self.MyDataClassA, 'a', default=self.MyDataClassA(), title='CustomA title')
+        parser.add_dataclass_arguments(self.MyDataClassB, 'b', default=self.MyDataClassB())
+
+        cfg = parser.get_defaults()
+        self.assertEqual(self.dataclasses.asdict(self.MyDataClassA()), cfg['a'].as_dict())
+        self.assertEqual(self.dataclasses.asdict(self.MyDataClassB()), cfg['b'].as_dict())
+        dump = yaml.safe_load(parser.dump(cfg))
+        self.assertEqual(self.dataclasses.asdict(self.MyDataClassA()), dump['a'])
+        self.assertEqual(self.dataclasses.asdict(self.MyDataClassB()), dump['b'])
+
+        cfg = parser.instantiate_classes(cfg)
+        self.assertIsInstance(cfg['a'], self.MyDataClassA)
+        self.assertIsInstance(cfg['b'], self.MyDataClassB)
+        self.assertIsInstance(cfg['b'].b2, self.MyDataClassA)
+
+        self.assertEqual(5, parser.parse_args(['--b.b2.a1=5']).b.b2.a1)
+        self.assertRaises(ParserError, lambda: parser.parse_args(['--b.b2.a1=x']))
+
+        help_str = StringIO()
+        parser.print_help(help_str)
+        self.assertIn('CustomA title:', help_str.getvalue())
+        self.assertIn('MyDataClassB description:', help_str.getvalue())
+        self.assertIn('b2 help:', help_str.getvalue())
+
+        with self.assertRaises(ValueError):
+            parser.add_dataclass_arguments(1, 'c')
+
+        with self.assertRaises(ValueError):
+            parser.add_dataclass_arguments(self.MyDataClassB, 'c', default=self.MyDataClassB(b2=self.MyDataClassB()))
+
+        class MyClass(int, self.MyDataClassA):
+            """MyClass description"""
+
+        with self.assertRaises(ValueError):
+            parser.add_dataclass_arguments(MyClass, 'c')
+
+
+    def test_dataclass_typehint(self):
+
+        class MyClass:
+            def __init__(
+                self,
+                a1: self.MyDataClassA = self.MyDataClassA(),  # type: ignore
+                a2: self.MyDataClassB = self.MyDataClassB(),  # type: ignore
+            ):
+                self.a1 = a1
+                self.a2 = a2
+
+        parser = ArgumentParser()
+        parser.add_class_arguments(MyClass, 'g')
+
+        cfg = parser.get_defaults()
+        self.assertEqual(self.dataclasses.asdict(self.MyDataClassA()), cfg['g']['a1'].as_dict())
+        self.assertEqual(self.dataclasses.asdict(self.MyDataClassB()), cfg['g']['a2'].as_dict())
+        dump = yaml.safe_load(parser.dump(cfg))
+        self.assertEqual(self.dataclasses.asdict(self.MyDataClassA()), dump['g']['a1'])
+        self.assertEqual(self.dataclasses.asdict(self.MyDataClassB()), dump['g']['a2'])
+
+        cfg_init = parser.instantiate_classes(cfg)
+        self.assertIsInstance(cfg_init.g.a1, self.MyDataClassA)
+        self.assertIsInstance(cfg_init.g.a2, self.MyDataClassB)
+        self.assertIsInstance(cfg_init.g.a2.b2, self.MyDataClassA)
+
+
+    def test_dataclass_typehint_in_subclass(self):
+
+        class MyClass1:
+            def __init__(self, a1: self.MyDataClassB = self.MyDataClassB()):  # type: ignore
+                """MyClass1 description"""
+                self.a1 = a1
+
+        class MyClass2:
+            def __init__(self, c1: MyClass1):
+                """MyClass2 description"""
+                self.c1 = c1
+
+        parser = ArgumentParser(error_handler=None)
+        parser.add_class_arguments(MyClass2)
+
+        from jsonargparse_tests import signatures_tests
+        setattr(signatures_tests, 'MyClass1', MyClass1)
+
+        class_path = '"class_path": "jsonargparse_tests.signatures_tests.MyClass1"'
+        init_args = '"init_args": {"a1": {"b2": {"a1": 7}}}'
+        cfg = parser.parse_args(['--c1={'+class_path+', '+init_args+'}'])
+        self.assertEqual(cfg.c1.class_path, 'jsonargparse_tests.signatures_tests.MyClass1')
+        self.assertEqual(cfg.c1.init_args.a1.b2.a1, 7)
+        self.assertIsInstance(cfg.c1.init_args.a1.b2.a1, PositiveInt)
+        cfg = parser.instantiate_classes(cfg)
+        self.assertIsInstance(cfg['c1'], MyClass1)
+        self.assertIsInstance(cfg['c1'].a1, self.MyDataClassB)
+        self.assertIsInstance(cfg['c1'].a1.b2, self.MyDataClassA)
+        self.assertIsInstance(cfg['c1'].a1.b1, PositiveFloat)
+
+
+    def test_dataclass_add_argument_type(self):
+        parser = ArgumentParser()
+        parser.add_argument('--b', type=self.MyDataClassB, default=self.MyDataClassB(b1=7.0))
+
+        cfg = parser.get_defaults()
+        self.assertEqual({'b1': 7.0, 'b2': {'a1': 1, 'a2': '2'}}, cfg['b'].as_dict())
+
+        cfg = parser.instantiate_classes(cfg)
+        self.assertIsInstance(cfg['b'], self.MyDataClassB)
+        self.assertIsInstance(cfg['b'].b2, self.MyDataClassA)
+
+
+    def test_dataclass_add_argument_type_some_required(self):
+
+        @self.dataclasses.dataclass
+        class MyDataClass:
+            a1: str
+            a2: float = 1.2
+
+        parser = ArgumentParser(error_handler=None)
+        parser.add_argument('--b', type=MyDataClass)
+
+        self.assertEqual(Namespace(a1='v', a2=1.2), parser.parse_args(['--b.a1=v']).b)
+        self.assertRaises(ParserError, lambda: parser.parse_args([]))
+
+
+    def test_dataclass_field_default_factory(self):
+
+        @self.dataclasses.dataclass
+        class MyDataClass:
+            a1: List[int] = self.dataclasses.field(default_factory=lambda: [1, 2, 3])
+            a2: Dict[str, float] = self.dataclasses.field(default_factory=lambda: {'a': 1.2, 'b': 3.4})
+
+        parser = ArgumentParser()
+        parser.add_class_arguments(MyDataClass)
+
+        cfg = parser.get_defaults().as_dict()
+        self.assertEqual([1, 2, 3], cfg['a1'])
+        self.assertEqual({'a': 1.2, 'b': 3.4}, cfg['a2'])
+
+
+    def test_compose_dataclasses(self):
+        dataclasses = import_dataclasses('test_compose_dataclasses')
+
+        @dataclasses.dataclass
+        class MyDataClassA:
+            a: int = 1
+
+            def __post_init__(self):
+                self.a += 1
+
+        @dataclasses.dataclass
+        class MyDataClassB:
+            b: str = '1'
+
+        MyDataClassAB = compose_dataclasses(MyDataClassA, MyDataClassB)
+        self.assertEqual(2, len(dataclasses.fields(MyDataClassAB)))
+        self.assertEqual({'a': 3, 'b': '2'}, dataclasses.asdict(MyDataClassAB(a=2, b='2')))  # pylint: disable=unexpected-keyword-arg
+
+
+    def test_instantiate_classes_dataclasses_lightning_issue_9207(self):
+        # https://github.com/PyTorchLightning/pytorch-lightning/issues/9207
+        dataclasses = import_dataclasses('test_instantiate_classes_dataclasses_lightning_issue_9207')
+
+        @dataclasses.dataclass
+        class MyDataClass:
+            name: str = 'name'
+
+        class MyClass:
+            def __init__(self, data: MyDataClass):
+                self.data = data
+
+        parser = ArgumentParser()
+        parser.add_class_arguments(MyClass, 'class')
+        cfg = parser.parse_args([])
+        cfg = parser.instantiate_classes(cfg)
+        self.assertIsInstance(cfg['class'], MyClass)
+        self.assertIsInstance(cfg['class'].data, MyDataClass)
+
+
 if __name__ == '__main__':
     unittest.main(verbosity=2)
